#include <MT/array.h>
#include <fstream>
#include <math.h>

void save(const arr& x,const char *filename){
    std::ofstream os(filename);
    x.writeRaw(os);
    os <<std::endl;
}

void exercise(){
  uint T=40,K=40;
  double w=1.,eps=.0; //eps=0 <-> no observation noise
  uint i,j,t;
  double d,Z;
  
  //-- build transition and observation matrices
  arr P0(K),A(K,K),B(K,K);
  for(i=0;i<K;i++) for(j=0;j<i;j++){
    d=(double)i-j;
    A(i,j)=A(j,i)=::exp(-d*d/(2.*w*w));
    B(i,j)=B(j,i)=eps;
    //B(i,j)=B(j,i)=::exp(-d*d/(2.*w*w)); //Gaussian observations
  }
  for(i=0;i<K;i++){
    P0(i)=1./K;
    A(i,i)=1.;
    B(i,i)=1.;
  }
  for(i=0;i<K;i++){ //normalize columns
    Z=0.;  for(j=0;j<K;j++) Z+=A(j,i);  for(j=0;j<K;j++) A(j,i)/=Z;
    Z=0.;  for(j=0;j<K;j++) Z+=B(j,i);  for(j=0;j<K;j++) B(j,i)/=Z;
  }
  save(A,"z.A");
  save(B,"z.B");

  //-- evidences from observations
  arr phi(T+1,K);
  phi=1.;           //phi=1 corresponds to no observation
  phi[0]=B[0];      //first observation (at t=0)
  phi[T/2]=B[0];    //intermediate (at t=T/2)
  phi[T]=B[K-1];  //last observation (at t=T-1)
  save(phi,"z.phi");

  //-- fwd & bwd messages
  arr a(T+1,K),b(T+1,K);
  /* we use a[0] as an extra message that representes the factor P(x_0)
     and b[T] as an extra message that encodes no information (equal to 1)
     -- this simplifies the implementation of the loops below: */
  a[0]()=P0;
  for(t=1;t<=T;t++) a[t]() = A*(phi[t-1]%a[t-1]); // %=element-wise multiplication, *=inner product
  b[T]()=1.;
  for(t=T;t--;)  b[t]() = (phi[t+1]%b[t+1])*A;
  save(a,"z.alpha");
  save(b,"z.beta");

  //-- posterior beliefs
  arr post(T+1,K);
  for(t=0;t<=T;t++) post[t]() = a[t]%phi[t]%b[t];

  save(post,"z.post");
  
  //gnuplot: splot 'z.post' matrix with pm3d
}


int main(int argn,char** argv){
  exercise();

  return 0;
}
